import decimal

from measurement.base import AbstractMeasure, MetricUnit, Unit

__all__ = ["Time", "Frequency"]


class Time(AbstractMeasure):
    """
    Time measurements (generally for multidimensional measures).

    Please do not use this for handling durations of time unrelated to
    measure classes -- python's built-in datetime module has much better
    functionality for handling intervals of time than this class provides.
    """

    second = MetricUnit("1", ["s", "sec", "seconds"], ["s"], ["sec"])
    minute = Unit("60", ["min", "minutes"])
    hour = Unit("3600", ["hr", "h", "hours"])
    day = Unit("86400", ["d", "days"])
    julian_year = MetricUnit(
        "31557600",
        ["year", "a", "aj", "years", "annum", "Julian year"],
        ["a"],
        ["annum"],
    )


class Frequency(AbstractMeasure):
    hertz = MetricUnit("1", ["Hz", "Hertz"], ["Hz"], ["hertz"])
    rpm = Unit(decimal.Decimal("1.0") / decimal.Decimal("60"), ["RPM", "bpm", "BPM"])

    def __mul__(self, other):
        if isinstance(other, Time):
            return self.si_value * other.si_value
        return super().__mul__(other)



from measurement.base import AbstractMeasure, MetricUnit

__all__ = ["Radioactivity"]


class Radioactivity(AbstractMeasure):
    """Radioactivity measurements."""

    becquerel = MetricUnit("1", ["Bq"], ["Bq"])
    curie = MetricUnit("37000000000", ["Ci"], ["Ci"])
    rutherford = MetricUnit("1000000", ["Rd"], ["Rd"])



from measurement.base import AbstractMeasure, MetricUnit, Unit

__all__ = ["Energy", "Heat"]


class Energy(AbstractMeasure):
    joule = MetricUnit("1", ["J", "Joule"], ["J"], ["joule"])
    calorie = MetricUnit(
        "4184.0", ["c", "cal", "Cal", "Calorie", "C"], ["cal"], ["calorie"]
    )
    electronvolt = MetricUnit(
        "1.602177e-19",
        ["eV", "electron-volt", "electron volt"],
        ["eV"],
        ["electronvolt"],
    )
    tonne_tnt = Unit("4184000000")


Heat = Energy



import dataclasses
import decimal
from typing import List

from ..base import AbstractMeasure, AbstractUnit, MetricUnit

__all__ = ["Temperature"]


@dataclasses.dataclass
class DegreeUnit(AbstractUnit):
    symbols: List[str] = dataclasses.field(default_factory=list)
    """Symbols used to describe this unit."""

    def get_symbols(self):
        yield self.name.replace("_", " "), type(self)()
        yield from ((name, type(self)()) for name in self.symbols)


class DegreeCelcius(DegreeUnit):
    def to_si(self, value):
        return value + decimal.Decimal("273.15")

    def from_si(self, value):
        return value - decimal.Decimal("273.15")


class DegreeFahrenheit(DegreeUnit):
    def to_si(self, value):
        celsius = (value - 32) * 5 / 9
        return celsius + decimal.Decimal("273.15")

    def from_si(self, value):
        celsius = value - decimal.Decimal("273.15")
        return celsius * 9 / 5 + 32


class Temperature(AbstractMeasure):
    kelvin = MetricUnit("1", ["K", "Kelvin"], ["K"], ["kelvin"])
    celsius = DegreeCelcius(["°C"])
    fahrenheit = DegreeFahrenheit(["°F"])



def guess(value, unit, measures=None):
    """
    Return measurement instance based on given unit.

    Raises:
        ValueError: If measurement type cannot be guessed.

    Returns:
        MeasureBase: Measurement instance based on given unit.

    """
    from measurement.base import AbstractMeasure

    for measure in measures or AbstractMeasure.__subclasses__():
        try:
            return measure(**{unit: value})
        except KeyError:
            pass
    raise ValueError(f"can't guess measure for '{value} {unit}'")



from .electromagnetism import *  # NoQA
from .energy import *  # NoQA
from .geometry import *  # NoQA
from .mechanics import *  # NoQA
from .radioactivity import *  # NoQA
from .temperature import *  # NoQA
from .time import *  # NoQA



import decimal

from measurement.base import AbstractMeasure, MeasureBase, MetricUnit, Unit

from .geometry import Distance, Volume
from .time import Time

__all__ = ["Mass", "Pressure", "VolumetricFlowRate", "Speed"]


class FractionMeasureBase(MeasureBase):
    def __new__(mcs, name, bases, attrs):
        mcs.freeze_org_units(attrs)
        numerator = attrs["__numerator__"]
        denominator = attrs["__denominator__"]
        attrs.update(mcs.div(numerator, denominator))

        cls = super().__new__(mcs, name, bases, attrs)
        return cls

    @staticmethod
    def div(numerator, denominator):
        for numerator_name, numerator_unit in numerator._units.items():
            for (
                denominator_name,
                denominator_unit,
            ) in denominator._units.items():
                name = f"{numerator_name}/{denominator_name}"
                factor = numerator_unit.factor / denominator_unit.factor
                unit = Unit(factor=factor)
                unit.name = name
                yield name, unit


class VolumetricFlowRate(AbstractMeasure, metaclass=FractionMeasureBase):
    """Volumetric Flow measurements (generally for water flow)."""

    __numerator__ = Volume
    __denominator__ = Time

    cms = Unit("1", ["cumecs", "CMS"])
    cfs = Unit(1 / decimal.Decimal("35.31466672148859025043801035"), ["CFS"])

    miners_inch_50 = Unit("566.33693184e-6", ["mi-50"])
    miners_inch_40 = Unit("707.9211648e-6", ["mi-40"])
    miners_inch_38_4 = Unit("737.41788e-6", ["mi-38-4"])

    @classmethod
    def _attr_to_unit(cls, name):
        return super()._attr_to_unit(name.replace("__", "/"))


class Speed(AbstractMeasure, metaclass=FractionMeasureBase):
    __numerator__ = Distance
    __denominator__ = Time

    mph = Unit("0.44704")
    kph = Unit(1 / decimal.Decimal("3.6"))

    @classmethod
    def _attr_to_unit(cls, name):
        return super()._attr_to_unit(name.replace("__", "/"))


class Mass(AbstractMeasure):
    gram = MetricUnit("1", ["g", "Gram"], ["g"], ["gram"])
    tonne = Unit("1000000", ["t", "metric ton", "metric tonne"])
    ounce = Unit("28.34952", ["oz"])
    pound = Unit("453.59237", ["lb"])
    stone = Unit("6350.293")
    short_ton = Unit("907185.0", ["ton"])
    long_ton = Unit("1016000.0")


class Pressure(AbstractMeasure):
    pascal = MetricUnit("1", ["pa"], ["pa"], ["pascal"])
    bar = Unit("100000")
    atmosphere = Unit("101325", ["atm"])
    technical_atmosphere = Unit("98066.5", ["at"])
    torr = Unit("133.322")
    psi = Unit("6894.757293168", ["pounds per square inch"])



from measurement.base import AbstractMeasure, MetricUnit

__all__ = [
    "Capacitance",
    "Current",
    "ElectricPower",
    "Inductance",
    "Resistance",
    "Voltage",
]


class Capacitance(AbstractMeasure):
    farad = MetricUnit("1", ["F", "Farad"], ["F"], ["farad"])


class Current(AbstractMeasure):
    ampere = MetricUnit("1", ["A", "amp", "Ampere"], ["A"], ["ampere", "amp"])

    def __mul__(self, other):
        if isinstance(other, Voltage):
            return ElectricPower(self.si_value * other.si_value, "W")
        return super().__mul__(other)


class Resistance(AbstractMeasure):
    ohm = MetricUnit("1", ["Ohm", "Ω"], ["Ω"], ["ohm"])


class Voltage(AbstractMeasure):
    volt = MetricUnit("1", ["V", "Volt"], ["V"], ["volt"])

    def __mul__(self, other):
        if isinstance(other, Current):
            return ElectricPower(self.si_value * other.si_value, "W")
        return super().__mul__(other)


class Inductance(AbstractMeasure):
    henry = MetricUnit("1", ["H", "Henry"], ["H"], ["henry"])


class ElectricPower(AbstractMeasure):
    """
    Electric power is defined as :class:`Voltage` multiplied by :class:`Current`.

    This is why you can divide :class:`Current` to get the :class:`Voltage`
    by :class:`Voltage` to get the :class:`Current`
    or by :class:`Current` to get the :class:`Voltage`:

        >>> from measurement import measures
        >>> measures.ElectricPower('24 W') / measures.Voltage('12 V')
        Current(ampere="2")
        >>> measures.ElectricPower('24 W') / measures.Current('4 A')
        Voltage(volt="6")


    Analog to this, you can also multiply both :class:`Current`
    and :class:`Voltage` to get :class:`Current` to get the :class:`Voltage`:

        >>> from measurement import measures
        >>> measures.Current('2 A') * measures.Voltage('12 V')
        ElectricPower(watt="24")

    """

    watt = MetricUnit(
        "1", ["W", "VA", "Watt", "Voltampere"], ["W", "VA"], ["watt", "voltampere"]
    )

    def __truediv__(self, other):
        if isinstance(other, Current):
            return Voltage(volt=self.si_value / other.si_value)

        if isinstance(other, Voltage):
            return Current(ampere=self.si_value / other.si_value)

        return super().__truediv__(other)



import decimal

from measurement.base import AbstractMeasure, MeasureBase, MetricUnit, Unit

__all__ = ["Distance", "Area", "Volume"]


class Distance(AbstractMeasure):
    """
    Distance is a factor for both :class:`Area` and :class:`Volume`.

    If you multiply a :class:`Distance` with another :class:`Distance`,
    you will get a :class:`Area`:

        >>> from measurement import measures
        >>> measures.Distance('2 m') * measures.Distance('3 m')
        Area(metre²="6")

    If you multiply a :class:`Distance` with a :class:`Area` or two
    :class:`Distances<Distance>`, you will get a :class:`Volume`:

        >>> from measurement import measures
        >>> measures.Distance('2 m') * measures.Area('6 m²')
        Volume(metre³="12")

    You can also build the second and third power of a :class:`Distance`,
    to get a :class:`Area` or :class:`Volume`.

        >>> from measurement import measures
        >>> measures.Distance('2 m') ** 2
        Area(metre²="4")
        >>> measures.Distance('2 m') ** 3
        Volume(metre³="8")

    """

    metre = MetricUnit("1", ["m", "meter", "Meter", "Metre"], ["m"], ["metre", "meter"])
    parsec = MetricUnit("3.0857E+16", ["Parsec", "pc"], ["pc"], ["parsec"])
    astronomical_unit = MetricUnit(
        "1.495978707E+11", ["au", "ua", "AU"], ["au", "ua", "AU"]
    )
    foot = Unit("0.3048", ["ft", "feet", "Foot (International)"])
    inch = Unit("0.0254", ["in", "inches"])
    mile = Unit("1609.344", ["mi"])
    chain = Unit("20.1168")
    chain_benoit = Unit("20.116782", ["Chain (Benoit)"])
    chain_sears = Unit("20.1167645", ["Chain (Sears)"])
    british_chain_benoit = Unit("20.1167824944", ["British chain (Benoit 1895 B)"])
    british_chain_sears = Unit("20.1167651216", ["British chain (Sears 1922)"])
    british_chain_sears_truncated = Unit(
        "20.116756", ["British chain (Sears 1922 truncated)"]
    )
    british_foot = Unit(
        "0.304799471539", ["british_ft", "British foot", "British foot (Sears 1922)"]
    )
    british_yard = Unit(
        "0.914398414616", ["british_yd", "British yard", "British yard (Sears 1922)"]
    )
    clarke_ft = Unit("0.3047972654", ["Clarke's Foot"])
    clarke_link = Unit("0.201166195164", ["Clarke's link"])
    fathom = Unit("1.8288")
    german_meter = Unit("1.0000135965", ["german_m", "German legal metre"])
    gold_coast_foot = Unit("0.304799710181508", ["gold_coast_ft", "Gold Coast foot"])
    indian_yard = Unit("0.914398530744", ["indian_yd", "Indian yard", "Yard (Indian)"])
    link = Unit("0.201168", ["Link"])
    link_benoit = Unit("0.20116782", ["Link (Benoit)"])
    link_sears = Unit("0.20116765", ["Link (Sears)"])
    nautical_mile = Unit("1852", ["Nautical Mile", "NM", "nmi"])
    nautical_mile_uk = Unit("1853.184", ["nm_uk", "Nautical Mile (UK)"])
    rod = Unit("5.0292")
    sears_yard = Unit("0.91439841", ["sears_yd", "Yard (Sears"])
    survey_foot = Unit("0.304800609601", ["survey_ft", "US survey foot", "U.S. Foot"])
    yard = Unit("0.9144", ["yd"])
    point = Unit(decimal.Decimal("25.4e-3") / 72, ["pt"])

    def __mul__(self, other):
        if isinstance(other, Distance):
            return Area(sq_metre=self.si_value * other.si_value)

        if isinstance(other, Area):
            return Volume(cubic_metre=self.si_value * other.si_value)

        return super().__mul__(other)

    def __pow__(self, power, modulo=None):
        if power == 2:
            return self * self
        if power == 3:
            return self * self * self
        return NotImplemented


class AreaBase(MeasureBase):
    def __new__(mcs, name, bases, attrs):
        mcs.freeze_org_units(attrs)
        x, y = attrs["__factors__"]
        if x is y:
            attrs.update(mcs.square(x))

        cls = super().__new__(mcs, name, bases, attrs)
        return cls

    @staticmethod
    def square(klass):
        for name, unit in klass._units.items():
            qs_unit = Unit(factor=unit.factor**2)
            qs_unit.name = f"{qs_unit.name}²"
            yield f"{name}²", qs_unit


class Area(AbstractMeasure, metaclass=AreaBase):
    """
    An area is defined as :class:`Distance` multipled by :class:`Distance`.

    This is why you can multiply two :class:`Distances<Distance>` to get
    an :class:`Area` or divide an :class:`Area` to get a :class:`Distance`:

        >>> from measurement import measures
        >>> measures.Distance('2 m') * measures.Distance('3 m')
        Area(metre²="6")
        >>> measures.Area('6 m²') / measures.Distance('2 m')
        Distance(metre="3")

    If you multiply a :class:`Area` with a :class:`Distance`,
    you will get a :class:`Volume`:

        >>> measures.Area('6 m²') * measures.Distance('2 m')
        Volume(metre³="12")

    """

    __factors__ = (Distance, Distance)

    acre = Unit(
        (decimal.Decimal("43560") * (Distance(ft=decimal.Decimal(1)).m ** 2)),
        ["Acre"],
    )
    hectare = Unit("10000", ["Hectare", "ha"])

    @classmethod
    def _attr_to_unit(cls, name):
        if name[:3] in ["sq_", "sq "]:
            name = f"{name[3:]}²"
        elif name[:7] in ["square_", "square "]:
            name = f"{name[7:]}²"
        return super()._attr_to_unit(name)

    def __truediv__(self, other):
        if isinstance(other, Distance):
            return Distance(metre=self.si_value / other.si_value)

        return super().__truediv__(other)

    def __mul__(self, other):
        if isinstance(other, Distance):
            return Volume(cubic_metre=self.si_value * other.si_value)

        return super().__mul__(other)


class VolumeBase(MeasureBase):
    def __new__(mcs, name, bases, attrs):
        mcs.freeze_org_units(attrs)
        if "__factors__" in attrs:
            x, y, z = attrs["__factors__"]
            attrs.update(mcs.cubic(x))
        cls = super().__new__(mcs, name, bases, attrs)
        return cls

    @staticmethod
    def cubic(klass):
        for name, unit in klass._units.items():
            qs_unit = Unit(factor=unit.factor**3)
            qs_unit.name = f"{qs_unit.name}³"
            yield f"{name}³", qs_unit


class Volume(AbstractMeasure, metaclass=VolumeBase):
    """
    A volume is defined as :class:`Area` multipled by :class:`Distance`.

    This is why you can multiply three :class:`Distances<Distance>` to get
    a :class:`Volume`, multiply a :class:`Area` by a :class:`Distance`
    or divide a :class:`Volume` by both :class:`Distance` and :class:`Area`:

        >>> from measurement import measures
        >>> measures.Distance('2 m') * measures.Distance('3 m') * measures.Distance('4 m')
        Volume(metre³="24")
        >>> measures.Distance('2 m') * measures.Area('6 m²')
        Volume(metre³="12")
        >>> measures.Volume('12 m³') / measures.Area('6 m²')
        Distance(metre="2")
        >>> measures.Volume('12 m³') / measures.Distance('6 m')
        Area(metre²="2")

    """

    __factors__ = (Distance, Distance, Distance)

    litre = MetricUnit(
        "1e-3", ["liter", "L", "l", "ℓ"], ["L", "l", "ℓ"], ["litre", "liter"]
    )

    us_gallon = Unit(
        "3.785411784e-3", ["US gallon", "US gal", "US fluid gallon", "gallon (US)"]
    )
    us_fluid_ounce = Unit("29.57353e-6", ["US oz", "US fl oz"])
    us_fluid_ounce_food = Unit("30e-6", ["US fluid ounce (food nutrition labelling)"])
    us_liquid_quart = Unit("0.946352946e-3", ["US liquid quart"])
    us_liquid_pint = Unit("473.176473e-6")
    us_gill = Unit("118.29411825e-6")
    us_tablespoon = Unit("14.78676478125e-6", ["US tablespoon", "Us tbsp", "Us Tbsp"])
    us_tsp = Unit("4.92892159375e-6", ["US tsp"])
    us_fluid_darm = Unit("3.6966911953125e-6")

    us_dry_gallon = Unit(
        "4.40488377086e-3", ["US dry gallon", "corn gallon", "grain gallon"]
    )
    us_dry_quart = Unit("1.101220942715e-3")
    us_dry_pint = Unit("550.6104713575e-6")
    us_bushel = Unit("35.23907016688e-3", ["US bsh", "US bu"])

    cubic_inch = Unit("16.387064e-6", ["cu in"])
    cubic_foot = Unit("0.02832", ["cu ft"])

    imperial_gallon = Unit(
        "4.54609e-3", ["Imperial gallon", "Imperial gal", "gallon (Imperial)"]
    )
    imperial_fluid_ounce = Unit("28.41306e-6", ["Imperial fluid ounce", "imp fl oz"])
    imperial_quart = Unit("1.1365225e-6", ["Imperial quart"])
    imperial_pint = Unit("568.26125e-6", ["Imperial pint"])
    imperial_gill = Unit("142.0653125e-6", ["Imperial gill"])
    imperial_bushel = Unit("36.36872e-3", ["imp bsh", "imp bu"])
    imperial_fluid_darm = Unit("3.5516328125e-6", ["luid drachm"])

    au_tablespoon = Unit("20e-6", ["Australian tablespoon", "Australian tbsp"])
    au_teaspoon = Unit(
        "5e-6",
        [
            "US tsp (food nutrition labelling)",
            "metic teaspoon",
            "metric tsp",
            "Australian tsp",
        ],
    )

    oil_barrel = Unit("158.987294928e-3", ["oil bbl", "bbl"])

    acre_in = Unit(
        "102.790153128960",
        ["ac in", "ac-in", "acre inches", "acre inch", "acre-inches", "acre-inch"],
    )
    acre_ft = Unit(
        "1233.481837547520",
        ["ac ft", "ac-ft", "acre feet", "acre foot", "acre-feet", "acre-foot"],
    )

    @classmethod
    def _attr_to_unit(cls, name):
        if name[:6] in ["cubic_", "cubic "]:
            name = f"{name[6:]}³"
        return super()._attr_to_unit(name)

    def __truediv__(self, other):
        if isinstance(other, Distance):
            return Area(sq_metre=self.si_value / other.si_value)

        if isinstance(other, Area):
            return Distance(metre=self.si_value / other.si_value)

        return super().__truediv__(other)



"""Collection of helpers and base classes to build measure."""
import abc
import dataclasses
import decimal
import inspect
import warnings
from functools import total_ordering
from typing import Any, Dict, Iterable, List, Optional, Tuple, Type, Union


def qualname(obj: Any) -> str:
    return obj.__qualname__ if inspect.isclass(obj) else type(obj).__qualname__


class ImmutableKeyDict(Dict):
    """Like :class:`.dict` but any key may only assigned to a value once."""

    def __setitem__(self, key, value):
        """
        Map item to key once and raise error if the same key is set twice.

        Raises:
            KeyError: If key has been already assigned to a different item.
        """
        if key in self.keys() and value is not self[key]:
            raise KeyError(f"Key '{key}' already exists with value '{self[key]}'.")
        dict.__setitem__(self, key, value)


class AbstractUnit(abc.ABC):
    """
    Helper class to define units of measurement in relation to their SI definition.

    Nowadays all units of measurement are defined based on fundamental SI units.
    This class provides behavior to convert a SI unit based measure to another Unit.
    """

    name = None

    @abc.abstractmethod
    def to_si(self, value: decimal.Decimal) -> decimal.Decimal:
        """Return SI measure based on given value in the unit defined by this class."""

    @abc.abstractmethod
    def from_si(self, value: decimal.Decimal) -> decimal.Decimal:
        """Return measure in the unit defined by this class based on given SI measure."""

    @abc.abstractmethod
    def get_symbols(self) -> Iterable[Tuple[str, Type["AbstractUnit"]]]:
        """Return list of symbol names and their :class:`.AbstractUnit` representation."""

    def __str__(self):
        return str(self.name)


@dataclasses.dataclass
class Unit(AbstractUnit):
    """
    Helper class for units that are defined as a multiple of an SI unit.

    Usage::

        from measurements.base import AbstractMeasure, Unit


        class Distance(AbstractMeasure):
            metre = Unit("1", ["m", "meter"])
            inch = Unit("0.0254", ["in", "inches"])

    In the example above we implemented a simple distance measure. This first
    argument of a unit is it's SI unit factor. For the SI unit itself –
    in this example metre – it's "1". For inches, the factor is by which number
    you would need to multiple a meter to get to an inch. Or, 0.0254 metres make
    an inch.

    The second argument is a list of symbols, that are used to describe the unit.
    In the case of a metre, it is ``m`` and also the American English `meter`.
    Note that the attribute name itself, will also be added automatically to
    the list of symbols.

    In a real-world measure, meter in the example above would be defined as a
    MetricUnit, as demonstrated below, allowing it to use all of the metric prefixes.
    """

    factor: Union[str, decimal.Decimal] = None
    """
    Factor of given measure based on SI unit.

    The given value must be either a :class:`Decimal<decimal.Decimal>`
    or a string that can be used to construct a decimal.
    """

    symbols: List[str] = dataclasses.field(default_factory=list)
    """Symbols used to describe this unit."""

    def __post_init__(self):
        if self.factor is not None:
            self.factor = decimal.Decimal(self.factor)

    def to_si(self, value):
        return value * self.factor

    def from_si(self, value):
        return value / self.factor

    def get_symbols(self):
        yield self.name.replace("_", " "), Unit(self.factor)
        yield from ((name, Unit(self.factor)) for name in self.symbols)


@dataclasses.dataclass
class MetricUnit(Unit):
    """Like :class:`.Unit` but with metric prefixes like ``kilo`` for ``kilometre``.

    Usage::

        from measurements.base import AbstractMeasure, Unit


        class Distance(AbstractMeasure):
            metre = MetricUnit(
                "1", ["m", "meter", "Meter", "Metre"], ["m"], ["metre", "meter"]
            )
            inch = Unit("0.0254", ["in", "inches"])

    In the example above, we update our simple distance measure to use MetricUnit
    for the metric metre unit. As before, the first argument of a unit is it's SI
    unit factor. For the SI unit itself – in this example metre – it's "1".

    The second argument is a list of symbols, that are used to describe the unit.
    Note that the attribute name itself, will also be added automatically to
    the list of symbols.

    The third argument is the small_metric_symbol, a List of symbols that are used
    with single letter metric prefixes, such as ``m`` for ``km`` or ``μm``.

    The fourth argument is metric_prefix, a list of symbols that are used with full
    words metric prefixes, such as ``metre`` for ``kilometre``.
    """

    small_metric_symbol: List[str] = dataclasses.field(default_factory=list)
    metric_prefix: List[str] = dataclasses.field(default_factory=list)

    SI_PREFIXE_SYMBOLS = {
        "y": decimal.Decimal("1e-24"),
        "z": decimal.Decimal("1e-21"),
        "a": decimal.Decimal("1e-18"),
        "f": decimal.Decimal("1e-15"),
        "p": decimal.Decimal("1e-12"),
        "n": decimal.Decimal("1e-9"),
        "u": decimal.Decimal("1e-6"),
        "μ": decimal.Decimal("1e-6"),
        "m": decimal.Decimal("1e-3"),
        "c": decimal.Decimal("1e-2"),
        "d": decimal.Decimal("1e-1"),
        "da": decimal.Decimal("1e1"),
        "h": decimal.Decimal("1e2"),
        "k": decimal.Decimal("1e3"),
        "M": decimal.Decimal("1e6"),
        "G": decimal.Decimal("1e9"),
        "T": decimal.Decimal("1e12"),
        "P": decimal.Decimal("1e15"),
        "E": decimal.Decimal("1e18"),
        "Z": decimal.Decimal("1e21"),
        "Y": decimal.Decimal("1e24"),
    }

    SI_PREFIXES = {
        "yocto": decimal.Decimal("1e-24"),
        "zepto": decimal.Decimal("1e-21"),
        "atto": decimal.Decimal("1e-18"),
        "femto": decimal.Decimal("1e-15"),
        "pico": decimal.Decimal("1e-12"),
        "nano": decimal.Decimal("1e-9"),
        "micro": decimal.Decimal("1e-6"),
        "milli": decimal.Decimal("1e-3"),
        "centi": decimal.Decimal("1e-2"),
        "deci": decimal.Decimal("1e-1"),
        "deca": decimal.Decimal("1e1"),
        "hecto": decimal.Decimal("1e2"),
        "kilo": decimal.Decimal("1e3"),
        "mega": decimal.Decimal("1e6"),
        "giga": decimal.Decimal("1e9"),
        "tera": decimal.Decimal("1e12"),
        "peta": decimal.Decimal("1e15"),
        "exa": decimal.Decimal("1e18"),
        "zeta": decimal.Decimal("1e21"),
        "yotta": decimal.Decimal("1e24"),
    }

    def get_symbols(self):
        yield from super().get_symbols()
        yield from (
            (f"{prefix}{s}", Unit(factor=self.factor * factor))
            for prefix, factor in self.SI_PREFIXE_SYMBOLS.items()
            for s in self.small_metric_symbol
        )
        yield from (
            (f"{prefix}{s}", Unit(factor=self.factor * factor))
            for prefix, factor in self.SI_PREFIXES.items()
            for s in self.metric_prefix
        )
        yield from (
            (f"{prefix}{s}".title(), Unit(factor=self.factor * factor))
            for prefix, factor in self.SI_PREFIXES.items()
            for s in self.metric_prefix
        )


class MeasureBase(type):
    """
    Create Measure class by unpacking all symbols into a dictionary.

    Units can be added to measure by adding :class:`.AbstractUnit` instances as
    class attributes to a Measure. This metaclass removes to attributes and
    creates a dictionary mapping all sympbols to their corresponding
    :class:`.AbstractUnit` implementation.

    Raises:
        KeyError: If the same symbol is used for multiple units.

    """

    def __new__(mcs, name, bases, attrs):
        mcs.freeze_org_units(attrs)
        symbols = ImmutableKeyDict()
        new_attr = {}
        for attr_name, attr in attrs.items():
            if isinstance(attr, AbstractUnit):
                attr.name = attr_name
                for symbol, unit in attr.get_symbols():
                    unit.name = attr_name
                    symbols[symbol] = unit
            else:
                new_attr[attr_name] = attr

        cls = super().__new__(mcs, name, bases, new_attr)
        cls._units = symbols
        return cls

    @staticmethod
    def freeze_org_units(attrs: Dict[str, Any]):
        if "_org_units" in attrs:
            return

        attrs["_org_units"] = {
            attr_name: attr
            for attr_name, attr in attrs.items()
            if isinstance(attr, Unit)
        }


@total_ordering
class AbstractMeasure(metaclass=MeasureBase):
    """Abstract super class to all measures."""

    def __init__(
        self,
        value: Union[str, decimal.Decimal, int, None] = None,
        unit: Optional[str] = None,
        **kwargs: Union[str, decimal.Decimal, int, None],
    ):
        if kwargs:
            unit, value = kwargs.popitem()
        if unit is None:
            value, unit = value.split(maxsplit=1)
            value = decimal.Decimal(value)

        if isinstance(value, int):
            value = decimal.Decimal(str(value))

        if not isinstance(value, (decimal.Decimal, str, int)):
            warnings.warn(f"'value' expects type Decimal not {qualname(value)}")
        value = decimal.Decimal(value)

        unit = self._attr_to_unit(unit)
        self.unit = self._units[unit]
        self.unit.org_name = unit
        self.si_value = self.unit.to_si(value)
        self.base_unit_names = self.get_base_unit_names()

    @classmethod
    def get_base_unit_names(cls):
        """Return a list of unit names for units with a factor of 1 (base units)."""
        names_list = [
            unit_name
            for unit_name, unit in cls._units.items()
            if getattr(unit, "factor", None) is not None
            and unit.factor == decimal.Decimal("1")
        ]
        if names_list:
            return names_list

    def __getattr__(self, name):
        try:
            unit = self._units[self._attr_to_unit(name)]
        except KeyError as e:
            raise AttributeError(
                f"{qualname(self)} object has no attribute '{name}'"
            ) from e
        else:
            return unit.from_si(self.si_value)

    def __getitem__(self, item):
        try:
            unit = self._units[self._attr_to_unit(item)]
        except KeyError as e:
            raise KeyError(f"{qualname(self)} object has no key '{item}'") from e
        else:
            return unit.from_si(self.si_value)

    @classmethod
    def _attr_to_unit(cls, name: str) -> str:
        return name.replace("_", " ")

    unit = None
    """Return :class:`~Unit` initially given to construct the measure."""

    @property
    def _value(self) -> decimal.Decimal:
        """Return :class:`~Decimal` value of measure in the given :attr:`.unit`."""
        return getattr(self, self.unit.name)

    def __repr__(self):
        return f'{qualname(self)}({self.unit.name}="{getattr(self, self.unit.name)}")'

    def __str__(self):
        return f"{getattr(self, self.unit.org_name)} {self.unit.org_name}"

    def __format__(self, format_spec):
        decimal_format = getattr(self, self.unit.org_name).__format__(format_spec)
        return f"{decimal_format} {self.unit.org_name}"

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.si_value == other.si_value

    def __lt__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.si_value < other.si_value

    def __gt__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.si_value > other.si_value

    def __add__(self, other):
        if not isinstance(other, type(self)):
            raise TypeError(f"can't add type '{qualname(self)}' to '{qualname(other)}'")
        return type(self)(
            value=self._value + getattr(other, self.unit.name), unit=self.unit.name
        )

    def __iadd__(self, other):
        return self + other

    def __sub__(self, other):
        if not isinstance(other, type(self)):
            raise TypeError(
                f"can't subtract type '{qualname(other)}' from '{qualname(self)}'"
            )

        return type(self)(
            value=self._value - getattr(other, self.unit.name), unit=self.unit.name
        )

    def __isub__(self, other):
        return self - other

    def __mul__(self, other):
        try:
            value = getattr(self, self.unit.org_name) * other
            return type(self)(value=value, unit=self.unit.org_name)
        except TypeError as e:
            raise TypeError(
                f"can't multiply type '{qualname(self)}' and '{qualname(other)}'"
            ) from e

    def __imul__(self, other):
        return self * other

    def __rmul__(self, other):
        return self * other

    def __truediv__(self, other):
        if isinstance(other, type(self)):
            return self.si_value / other.si_value
        try:
            value = getattr(self, self.unit.org_name) / other
        except TypeError as e:
            raise TypeError(
                f"can't divide type '{qualname(self)}' by '{qualname(other)}'"
            ) from e
        else:
            return type(self)(value=value, unit=self.unit.org_name)

    def __itruediv__(self, other):
        return self / other

    def __rtruediv__(self, other):
        return self / other

    def __bool__(self):
        return bool(self.si_value)
